from .attachment import *
from .cross_mention import *
from .flags import *
from .message import *
from .message_application import *
from .message_interaction import *
from .preinstanced import *
from .utils import *

__all__ = (
    *attachment.__all__,
    *cross_mention.__all__,
    *flags.__all__,
    *message.__all__,
    *message_application.__all__,
    *message_interaction.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)


from ...utils.module_deprecation import deprecated_import


@deprecated_import
class MessageReference:
    """
    A cross guild reference used as a ``Message``'s `.cross_reference` at crosspost messages.
    
    This type is Deprecated and will be removed in 2023 January.
    """
    __slots__ = ()


@deprecated_import
class MessageRepr:
    """
    Represents an uncached message.
    
    This type is Deprecated and will be removed in 2023 January.
    """
    __slots__ = ()



__all__ = ('parse_oauth2_redirect_url',)

import re

from .preinstanced import Oauth2Scope


OAUTH2_REQUEST_URL_RP = re.compile('(https?://.+?)\?code=([a-zA-Z0-9]{30})')


def parse_oauth2_redirect_url(url):
    """
    Parses the `redirect_url` and the `code` out from a whole `url`, what an user was redirected to after oauth2
    authorization.
    
    If the parsing was successful, then returns a `tuple` of `redirect_url` and `code`. If it fails, returns `None`.
    
    Parameters
    ----------
    url : `str`
        A whole to url to parse from

    Returns
    -------
    result : `None`, `tuple` (`str`, `str`)
    """
    result = OAUTH2_REQUEST_URL_RP.fullmatch(url)
    if result is None:
        return None
    
    return result.groups()


def parse_joined_oauth2_scopes(joined_scopes):
    """
    Parses the joined oauth2 scopes.
    
    Parameters
    ----------
    joined_scopes : `str`
        The joined scopes.
    
    Returns
    -------
    scopes : `None`, `tuple` of ``Oauth2Scope``
    """
    split_scopes = joined_scopes.split()
    if not split_scopes:
        return None
    
    split_scopes.sort()
    
    return tuple(Oauth2Scope.get(scope) for scope in split_scopes)


def parse_oauth2_scope_array(scope_array):
    """
    Parsers the given oauth2 scope array.
    
    Parameters
    ----------
    scope_array : `None`, `list` of `str`
        Scope array.
    
    
    Returns
    -------
    scopes : `None`, `tuple` of ``Oauth2Scope``
    """
    if (scope_array is None) or (not scope_array):
        return None
    
    scope_array.sort()
    
    return tuple(Oauth2Scope.get(scope) for scope in scope_array)


def join_oauth2_scopes(scopes):
    """
    Joins the given oauth2 scopes together.
    
    Parameters
    ----------
    scopes : `None`, `tuple` of ``Oauth2Scope``
        Oauth2 scopes to join.
    
    Returns
    -------
    joined_scopes : `str`
    """
    if scopes is None:
        return ''
    
    return ' '.join([scope.value for scope in scopes])


def make_oauth2_scope_array(scopes):
    """
    Converts the given oath2 scopes into a string array.
    
    Parameters
    ----------
    scopes : `None`, `tuple` of ``Oauth2Scope``
        Oauth2 scopes to use.
    
    Returns
    -------
    scope_array : `list` of `str`
    """
    if scopes is None:
        scope_array = []
    else:
        scope_array = [scope.value for scope in scopes]
    
    return scope_array


def _get_one_scope_value(scope):
    """
    Gets the value of one scope.
    
    Used by ``build_joined_scopes``.
    
    Returns
    -------
    scope_value : `str`
    
    Raises
    ------
    TypeError
        - If `scope` is neither ``Oauth2Scope``, `str`.
    """
    if isinstance(scope, Oauth2Scope):
        scope_value =  scope.value
    
    elif isinstance(scope, str):
        scope_value = scope
    
    else:
        raise TypeError(
            f'A scope can be `{Oauth2Scope.__name__}`, `str`; got {scope.__class__.__name__}; {scope!r}.'
        )
    
    return scope_value


def build_joined_scopes(scopes):
    """
    Builds joined scopes from the given value.
    
    Parameters
    ----------
    scopes : ``Oauth2Scope``, `str`, `iterable` of (``Oauth2Scope``, `str`)
        The scopes to join.
    
    Returns
    -------
    joined_scopes : `str`
    
    Raises
    ------
    TypeError
        - If `scopes`'s type is incorrect.
    """
    if isinstance(scopes, Oauth2Scope):
        joined_scopes = scopes.value
    
    elif isinstance(scopes, str):
        joined_scopes = scopes
    
    elif getattr(scopes, '__iter__', None) is not None:
        joined_scopes = ' '.join([_get_one_scope_value(scope) for scope in scopes])
    
    else:
        raise TypeError(
            f'`scopes` can be `str`, `{Oauth2Scope.__name__}`, `iterable` of them; '
            f'got {scopes.__class__.__name__}; {scopes!r}.'
        )
    
    return joined_scopes



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, force_string_parser_factory, preinstanced_array_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, default_entity_putter_factory, force_string_putter_factory,
    preinstanced_array_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator, force_string_validator_factory,
    preinstanced_array_validator_factory
)
from ...oauth2 import Oauth2Scope
from ...preconverters import preconvert_preinstanced_type
from ...user import ClientUserBase, User, ZEROUSER

from .constants import  NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import IntegrationType


# enabled

parse_enabled = bool_parser_factory('enabled', True)
put_enabled_into = bool_optional_putter_factory('enabled', True)
validate_enabled = bool_validator_factory('enabled')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# scopes

parse_scopes = preinstanced_array_parser_factory('scopes', Oauth2Scope)
put_scopes_into = preinstanced_array_putter_factory('scopes')
validate_scopes = preinstanced_array_validator_factory('scopes', Oauth2Scope)

# type

def parse_type(data):
    """
    Parses out an integration's type the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Entity data.
    
    Returns
    -------
    preinstanced : ``IntegrationType``
    """
    try:
        value = data['type']
    except KeyError:
        preinstanced = IntegrationType.none
    else:
        preinstanced = IntegrationType.get(value)
    
    return preinstanced


def put_type_into(preinstanced, data, defaults):
    """
    Puts the integration type the given `data` json serializable object.
    
    Parameters
    ----------
    preinstanced : ``IntegrationType``
        An integration's type.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    data['type'] = preinstanced.value
    
    return data


def validate_type(preinstanced):
    """
    Validates the given integration type.
    
    Parameters
    ----------
    preinstanced : `None`, ``IntegrationType``, `str`
        An integration's type.
    
    Returns
    -------
    preinstanced : ``IntegrationType``
    
    Raises
    ------
    TypeError
        - If `preinstanced`'s type is incorrect.
    """
    if preinstanced is None:
        return IntegrationType.none
    
    return preconvert_preinstanced_type(preinstanced, 'type', IntegrationType)

# user

parse_user = default_entity_parser_factory('user', User, ZEROUSER)
put_user_into = default_entity_putter_factory('user', ClientUserBase, ZEROUSER)
validate_user = default_entity_validator('user', ClientUserBase, ZEROUSER)
